//
// this file, part of libTexFormula, is released under the conditions of the zlib license.
// for details see: http://www.gzip.org/zlib/zlib_license.html
//
#include <stdio.h>
#include <stdlib.h>
#include "math.h"
#include "string.h"

#include <string>
#include <fstream>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN 1
	#include <windows.h>
	#include <direct.h>
	#define DIR_SEPERATOR "\\"
#else
	#include <fcntl.h>
	#include <unistd.h>
	#define max std::max
	#define min std::min
	#define DIR_SEPERATOR "/"
#endif

#include "libTeXFormula.h"
#include "libTeXFormulaInternal.h"
#include "display.h"
#include <algorithm>

extern int g_result_width, g_result_height;
extern unsigned char *g_result_pixmap;

static int g_resolution = 160;
static std::string tempDir;
static gs_main_instance *g_minst = NULL;
static bool g_gsLibAvailable = false;
static std::string g_gsCmdName;
static int g_AntialiasFactor = 4;

static display_callback display =
    {
        sizeof(display_callback),
        DISPLAY_VERSION_MAJOR,
        DISPLAY_VERSION_MINOR,
        display_open,
        display_preclose,
        display_close,
        display_presize,
        display_size,
        display_sync,
        display_page,
        display_update,
        NULL,    /* memalloc */
        NULL    /* memfree */
    };

// initializes the libTex Library
int DLLEXPORT initLibTeXFormula()
{

    // check if LaTeX is available
    int result = startProcess("pdflatex -version");

    if (result < 0)
        return LIBTEXF_ERROR_SYSTEM;
    if (result > 0)
        return LIBTEXF_ERROR_LATEXINIT;

    // create Temporary Directory
    tempDir = CreateTempDir();
    if (tempDir == "")
        return LIBTEXF_ERROR_NOTEMPDIR;


    // loads GhostScript Library
    if (LoadGSLibrary())
    {
        g_gsLibAvailable = true;

        // create ghostscript instance
        if (gsapi_new_instance_runtime(&g_minst, NULL) < 0 ||
                gsapi_set_display_callback_runtime(g_minst, &display) < 0)
        {
            freeLibTeXFormula();
            return LIBTEXF_ERROR_GSINIT;
        }
    }

    if (startProcess("gswin32c -v") == 0)
        g_gsCmdName = "gswin32c";
    else if (startProcess("gswin32 -v") == 0)
        g_gsCmdName = "gswin32";
    else if (startProcess("gs -v") == 0)
        g_gsCmdName = "gs";
    else if (!g_gsLibAvailable)
        return LIBTEXF_ERROR_GSINIT;

    return 1;
};

// Frees allocated resources
void DLLEXPORT freeLibTeXFormula()
{
    // free Temp Dir
    rmdir(tempDir.c_str());

    // release ghostscript instance
    if (g_minst)
        gsapi_delete_instance_runtime(g_minst);

    // free Ghostscript Libraray
    FreeGSLibrary();
};

// deletes temporary stuff
void cleanUp()
{
    unlink((tempDir + DIR_SEPERATOR "formula.pdf").c_str());
    unlink((tempDir + DIR_SEPERATOR "formula.aux").c_str());
    unlink((tempDir + DIR_SEPERATOR "formula.log").c_str());
    unlink((tempDir + DIR_SEPERATOR "formula.tex").c_str());
    unlink((tempDir + DIR_SEPERATOR "formula.pix").c_str());
};

// set resolution of output image
void DLLEXPORT setResolutionTeXFormula(int resolution)
{
    g_resolution = resolution;
};

// scales a image down using a simple linear filter
void antialiasImage(Image *theImage)
{
    // calculate new image size
    int newWidth =  (int) ceil((double) theImage->width / g_AntialiasFactor / 4.0) * 4,
                    newHeight = (int) ceil((double) theImage->height / g_AntialiasFactor);

    // allocate memory
    unsigned char *newImage = (unsigned char*) malloc(newWidth * newHeight);

    // scale image down
    for (int x = 0; x < newWidth;x++)
        for (int y = 0; y < newHeight;y++)
        {
            unsigned int sum = 0;

            // sum up intensities
            for (int x2 = x * g_AntialiasFactor; x2 < (x + 1) * g_AntialiasFactor;x2++)
                for (int y2 = y * g_AntialiasFactor; y2 < (y + 1) * g_AntialiasFactor;y2++)
                {
                    int imgX = max(0,min(theImage->width -1,x2)),
                               imgY = max(0,min(theImage->height -1,y2));
                    sum += theImage->pixmap[imgY * theImage->width + imgX];
                };

            // store new intensity
            sum /= (g_AntialiasFactor * g_AntialiasFactor);
            newImage[y * newWidth +x] = sum;

            // No gamma at the moment ...
            /*sum = sqrt((double) sum*sum*sum);
            sum /= 256;*/
            //double sumf = (double) sum / 256.0;
            //sum = pow(sumf, 1.8) * 256;//(double) sqrt((double) sum*sum*sum) * 256;
            //newImage[y * newWidth +x] = sum / (g_AntialiasFactor * g_AntialiasFactor));
        }

    // store new image in struct
    free(theImage->pixmap);
    theImage->pixmap = newImage;
    theImage->width = newWidth;
    theImage->height = newHeight;
};


// the forbidden commands
const char *unsecureCommands[] =
    {
        "\\catcode",
        "\\include",
        "\\includeonly",
        "\\input",
        "\\newcommand",
        "\\newenvironment",
        "\\newtheorem",
        "\\newfont",
        "\\renewcommand",
        "\\renewenvironment",
        "\\typein",
        "\\typeout",
        "\\write",
        NULL
    };

// checks if a string doesn't contain any unsecure LaTeX commands
//
// WARNING: We can't say for sure, that there are no other unsecure LaTeX commands.
//          The ones here are from wikitex, but there could still be other ones.
//          Using this LaTeX plugin might be a security risk!
//
bool isStringSecure(const char* string)
{
    // convert string to lower case
    std::string stringLower = string;
    std::transform(stringLower.begin(),stringLower.end(),stringLower.begin(),tolower);
    //char *stringLower = new char[strlen(string) + 1];
    //strcpy(stringLower,string);
    //strlwr(stringLower);

    // check string for forbidden commands
    for (const char **ptr = unsecureCommands; *ptr != NULL; ptr++)
        if (stringLower.find(*ptr) != std::string::npos)
            return false;

    // prevent user from leaving math mode
    bool escaped = false;
    for (unsigned int i = 0; i < strlen(string); i++)
    {
        switch (string[i])
        {
        case '\\':
            escaped = !escaped;
            break;
        case '$':
            if (!escaped)
                return false;
            escaped = false;
            break;
        default:
            escaped = false;
        };
    };

    return true;
};

// creates an image from an expression in LaTex
int DLLEXPORT parseTeXFormula(const char *formula, Image *theImage)
{
    // check if formula is secure
    if (!isStringSecure(formula))
        return LIBTEXF_ERROR_INSECURE_FORMULA;

    // create TeX File
    std::ofstream file((tempDir + DIR_SEPERATOR "formula.tex").c_str());
    if(!file)
        return LIBTEXF_ERROR_TEXFILE_OPEN;

    file    << "\\documentclass{minimal} "
    << "\\batchmode \\begin{document} "
    << "\\newcommand*{\\shipoutpdf}[1]{ "
    << "\\newpage "
    << "\\setlength{\\parindent}{0cm} "
    << "\\setlength{\\leftmargin}{0cm} "
    << "\\setbox0=\\vbox{\\kern -1truein \\kern .0ex"
    << "\\hbox{\\kern -1truein \\kern .0ex #1 \\kern 1truein \\kern -.7ex}"
    << "\\kern 1truein \\kern .0ex} "
    << "\\dimen0\\ht0 "
    << "\\advance\\dimen0\\dp0 "
    << "\\setlength{\\pdfpageheight}{\\dimen0} "
    << "\\setlength{\\pdfpagewidth}{\\wd0} "
    << "\\shipout\\box0 "
    << "} "
    << "\\shipoutpdf{$"<< formula <<"$} "
    << "\\end{document}\n";
    file.close();

    // create pdf from TeX File
    if (startProcess( "pdflatex formula.tex", tempDir.c_str() ) )
    {
        cleanUp();
        return LIBTEXF_ERROR_LATEXRUN;
    };

    // read pdf in ghostscript
    char * gsargv[12];
    char arg1[64];
    char arg2[64];
    sprintf(arg1, "-dDisplayHandle=%d", 0);
    sprintf(arg2, "-dDisplayFormat=%d",
            DISPLAY_COLORS_GRAY | DISPLAY_ALPHA_NONE | DISPLAY_DEPTH_8 |
            DISPLAY_LITTLEENDIAN);

    int gsargc;
    char filename[1024], res[15],res2[30];
    strcpy(filename,(tempDir + DIR_SEPERATOR "formula.pdf").c_str());
    sprintf(res,"-r%d",g_resolution * g_AntialiasFactor);

    // use the library
    if (g_gsLibAvailable)
    {
        gsargv[0] = "gs";	/* actual value doesn't matter */
        gsargv[1] = "-dNOPAUSE";
        gsargv[2] = "-dBATCH";
        gsargv[3] = "-dSAFER";
        gsargv[4] = "-sDEVICE=display";
        gsargv[5] = res;
        gsargv[6] = "-dTextAlphaBits=1";
        gsargv[7] = arg1;
        gsargv[8] = arg2;
        gsargv[9] = "-q";
        gsargv[10] = "-f";
        gsargv[11] = filename;
        gsargc=	12;
        if(gsapi_init_with_args_runtime(g_minst, gsargc, gsargv) < 0 || gsapi_exit_runtime(g_minst) < 0)
        {
            cleanUp();
            return LIBTEXF_ERROR_GSRUN;
        };

        theImage->width = g_result_width;
        theImage->height = g_result_height;
        theImage->pixmap = g_result_pixmap;

    }
    // use the command line tool
    else
    {
        // determine the image size
        std::string output;
        if (!startProcess( (g_gsCmdName + " -dNOPAUSE -dBATCH -dWhiteIsOpaque -dTextAlphaBits=1 -dSAFER -sDEVICE=bbox -q " +res +" -f formula.pdf").c_str(), tempDir.c_str(), NULL, &output ) == 0)
        {
            cleanUp();
            return LIBTEXF_ERROR_GSRUN;
        };

        float x1,x2,y1,y2;
        size_t pos = output.find("%%HiResBoundingBox");
        sscanf(output.c_str() + pos,"%%%%HiResBoundingBox: %f %f %f %f",&x1,&y1,&x2,&y2);

        theImage->width  = (int) ceil(x2 / 72.0 * g_resolution / 4) * 4 * g_AntialiasFactor,
                           theImage->height = (int) ceil(y2 / 72.0 * g_resolution ) * g_AntialiasFactor;

        // create the image
        sprintf(res2," -g%dx%d", theImage->width, theImage->height);
        if (!startProcess( (g_gsCmdName + " -dNOPAUSE -dBATCH -dSAFER -sDEVICE=bit -dTextAlphaBits=1 -dGrayValues=256 -q " +res + res2+ " -sOutputFile=formula.pix -f formula.pdf").c_str(), tempDir.c_str(), NULL, &output ) == 0)
        {
            cleanUp();
            return LIBTEXF_ERROR_GSRUN;
        };

        // read the image
        theImage->pixmap = (unsigned char*) malloc(theImage->width * theImage->height);
        std::ifstream image((tempDir + DIR_SEPERATOR "formula.pix").c_str());
        image.read((char*) theImage->pixmap, theImage->width * theImage->height);
        if (image.fail())
        {
            cleanUp();
            return LIBTEXF_ERROR_GSRUN;
        };

    };

    if (g_AntialiasFactor > 1)
        antialiasImage(theImage);

    // we tidy up behind ourselves! (at most times...)
    cleanUp();
    return 1;
};


// frees the memory used by an image
void DLLEXPORT freeTeXFormulaImage(Image *theImage)
{
    if (theImage->pixmap != NULL)
        free(theImage->pixmap);
};


